# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer/about.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(451, 334)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/app.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        AboutDialog.setWindowIcon(icon)
        self._central_layout = QtWidgets.QGridLayout(AboutDialog)
        self._central_layout.setObjectName("_central_layout")
        self.icon = QtWidgets.QLabel(AboutDialog)
        self.icon.setPixmap(QtGui.QPixmap(":/icons/app.png"))
        self.icon.setObjectName("icon")
        self._central_layout.addWidget(self.icon, 0, 0, 1, 1)
        self.title = QtWidgets.QLabel(AboutDialog)
        self.title.setObjectName("title")
        self._central_layout.addWidget(self.title, 0, 1, 1, 1)
        self.information = QtWidgets.QTabWidget(AboutDialog)
        self.information.setObjectName("information")
        self.detailsTab = QtWidgets.QWidget()
        self.detailsTab.setObjectName("detailsTab")
        self._details_layout = QtWidgets.QGridLayout(self.detailsTab)
        self._details_layout.setObjectName("_details_layout")
        self.details = QtWidgets.QLabel(self.detailsTab)
        self.details.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByKeyboard|QtCore.Qt.LinksAccessibleByMouse)
        self.details.setObjectName("details")
        self._details_layout.addWidget(self.details, 0, 0, 1, 1)
        self.information.addTab(self.detailsTab, "")
        self.licenseTab = QtWidgets.QWidget()
        self.licenseTab.setObjectName("licenseTab")
        self._license_layout = QtWidgets.QGridLayout(self.licenseTab)
        self._license_layout.setContentsMargins(0, 0, 0, 0)
        self._license_layout.setObjectName("_license_layout")
        self.license = QtWidgets.QPlainTextEdit(self.licenseTab)
        self.license.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.license.setReadOnly(True)
        self.license.setObjectName("license")
        self._license_layout.addWidget(self.license, 0, 0, 1, 1)
        self.information.addTab(self.licenseTab, "")
        self._central_layout.addWidget(self.information, 1, 0, 1, 2)
        self.buttons = QtWidgets.QDialogButtonBox(AboutDialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.buttons.setObjectName("buttons")
        self._central_layout.addWidget(self.buttons, 2, 0, 1, 2)
        self._central_layout.setColumnStretch(0, 1)
        self._central_layout.setColumnStretch(1, 4)
        self._central_layout.setRowStretch(0, 1)
        self._central_layout.setRowStretch(1, 5)

        self.retranslateUi(AboutDialog)
        self.buttons.accepted.connect(AboutDialog.accept) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        _translate = QtCore.QCoreApplication.translate
        self.information.setTabText(self.information.indexOf(self.detailsTab), _translate("AboutDialog", "Details"))
        self.information.setTabText(self.information.indexOf(self.licenseTab), _translate("AboutDialog", "License"))
from . import resources_rc
