# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer/help.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_HelpDialog(object):
    def setupUi(self, HelpDialog):
        HelpDialog.setObjectName("HelpDialog")
        HelpDialog.resize(760, 580)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/app.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        HelpDialog.setWindowIcon(icon)
        self.gridLayout = QtWidgets.QGridLayout(HelpDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.browser = QtWidgets.QTextBrowser(HelpDialog)
        self.browser.setOpenLinks(False)
        self.browser.setObjectName("browser")
        self.gridLayout.addWidget(self.browser, 0, 0, 1, 1)
        self.buttons = QtWidgets.QDialogButtonBox(HelpDialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttons.setObjectName("buttons")
        self.gridLayout.addWidget(self.buttons, 1, 0, 1, 1)

        self.retranslateUi(HelpDialog)
        self.buttons.rejected.connect(HelpDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(HelpDialog)

    def retranslateUi(self, HelpDialog):
        pass
from . import resources_rc
